use std::collections::HashMap;

pub trait Render<'a> {
    /// Render into HTML.
    fn render(
        self,
        plugins: &mut HashMap<String, Box<dyn FnMut(Element) -> Element + 'a>>,
    ) -> String;

    /// Render into HTML with no plugins.
    fn render_safe(self) -> String;
}

pub const SELF_CLOSING_TAGS: &[&str] = &["img", "br", "hr", "input", "meta", "link"];

/// The entire document can be boiled down to just elements with children... so just
/// a simple tree structure (just without the linking to parent since that's annoying)
#[derive(Clone, Debug, Default)]
pub struct Element {
    pub tag: String,
    pub attrs: HashMap<String, String>,
    pub children: Vec<Element>,
}

impl<'a> Render<'a> for Element {
    fn render(
        self,
        plugins: &mut HashMap<String, Box<dyn FnMut(Element) -> Element + 'a>>,
    ) -> String {
        if self.tag == "text" {
            return self.attrs.get("content").unwrap().to_owned();
        } else if self.tag == "v" {
            return self.attrs.get("0").unwrap().to_owned();
        } else if self.tag == "null?" {
            // just render children of null element
            let mut inner: String = String::new();

            for element in &self.children {
                inner.push_str(&element.clone().render(plugins));
            }

            return inner;
        } else if let Some(f) = plugins.get_mut(&self.tag) {
            return f(self).render(plugins);
        }

        let closing = format!("</{}>", self.tag);
        format!(
            "<{}{}{}{}{}",
            self.tag,
            {
                let mut attrs: String = String::new();

                for attr in &self.attrs {
                    attrs.push_str(" ");
                    attrs.push_str(attr.0);
                    attrs.push_str("=\"");
                    attrs.push_str(attr.1);
                    attrs.push_str("\"");
                }

                attrs
            },
            if SELF_CLOSING_TAGS.contains(&self.tag.as_str()) {
                "/>"
            } else {
                ">"
            },
            {
                let mut inner: String = String::new();

                for element in &self.children {
                    inner.push_str(&element.clone().render(plugins));
                }

                inner
            },
            if SELF_CLOSING_TAGS.contains(&self.tag.as_str()) {
                ""
            } else {
                &closing
            }
        )
    }

    fn render_safe(self) -> String {
        self.render(&mut {
            let mut h = HashMap::new();
            h.insert(
                "#null".to_string(),
                Box::new(|x: Element| x.to_owned()) as _,
            );
            h
        })
    }
}
